(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Critical_Pair_Closure_Impl
imports 
  Critical_Pair_Closure
  Critical_Pairs_Impl
  QDP_Framework_Impl
begin

definition check_strongly_closed where 
  "check_strongly_closed R n = do {
           check_linear_trs R;
           check_allm (\<lambda> (b,s,t). do {
              check (\<not> Option.is_none (List.find (\<lambda> x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0))) \<and>
                \<not> Option.is_none (List.find (\<lambda> x. List.member (reachable_terms R t n) x) (reachable_terms R s (Suc 0))))
               (shows ''there is a critical pair that is not strongly closed: '' +@+ shows_term s +@+ shows '' <- . -> ''
                +@+ shows_term t)
            }) (critical_pairs_impl R R)
         }  <+? (\<lambda>s. s +@+ shows_nl +@+ shows ''hence, the following TRS is not strongly closed '' +@+ shows_nl +@+ shows_trs R
             +@+ indent s)"

lemma check_strongly_closed: assumes "isOK(check_strongly_closed R n)"
  shows "CR (rstep (set R))"
proof (rule strongly_closed_linear_CR)
  from assms show "linear_trs (set R)" unfolding check_strongly_closed_def by auto
  {
    fix s t
    assume "\<not> Option.is_none (List.find (\<lambda> x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0)))"
    then obtain r where r:"(List.find (\<lambda> x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0))) = Some r" 
      unfolding is_none_def by auto
    hence "List.member (reachable_terms R s n) r" using find_Some_iff by metis
    hence 1:"r \<in> set (reachable_terms R s n)" using in_set_member by fast
    from r have "\<exists>i<length (reachable_terms R t (Suc 0)). r = (reachable_terms R t (Suc 0)) ! i" using find_Some_iff by metis
    hence "r \<in> set (reachable_terms R t (Suc 0))" using nth_mem by blast
    with 1 have "\<exists>r. (s,r) \<in> (rstep (set R))^* \<and> (t,r) \<in> (rstep (set R))^=" using reachable_terms reachable_terms_one by metis
  } note sound=this
  {
    fix b s t
    assume "(b,s,t) \<in> critical_pairs (set R) (set R)" 
    with assms have 1:"\<not> Option.is_none (List.find (\<lambda> x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0))) \<and>
    \<not> Option.is_none (List.find (\<lambda> x. List.member (reachable_terms R t n) x) (reachable_terms R s (Suc 0)))" 
      unfolding check_strongly_closed_def by fastforce
    with sound have "\<exists>r. (s,r) \<in> (rstep (set R))^* \<and> (t,r) \<in> (rstep (set R))^=" by blast
    moreover from sound 1 have "\<exists>u. (t,u) \<in> (rstep (set R))^* \<and> (s,u) \<in> (rstep (set R))^=" by blast
    ultimately have "\<exists>r u. (s, r) \<in> (rstep (set R))^* \<and> (t, r) \<in> (rstep (set R))^= 
      \<and> (s, u) \<in> (rstep (set R))^= \<and> (t, u) \<in> (rstep (set R))^*" by auto
  }
  thus "strongly_closed (set R)" unfolding strongly_closed_def by blast
qed

end