(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

header {* The Critical Pair Lemma (with variables fixed to type string) *}

theory Critical_Pairs
imports 
  Q_Restricted_Rewriting
  Unification
begin

(* we also consider overlaps between the same rule at top level,
   in this way we are not restricted to wf_trss *)
(* we use strings as variables for convenient working with variable renamings *)
definition
  critical_pairs :: "('f, string) trs \<Rightarrow> ('f, string) trs \<Rightarrow> (bool \<times> ('f, string) rule) set" 
where
  "critical_pairs P R = { (C = \<box>, r \<cdot> \<sigma>, (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>) | l r l' r' l'' C \<sigma> \<tau>. 
    (l, r) \<in> P \<and> (l', r') \<in> R \<and> l = C\<langle>l''\<rangle> \<and> is_Fun l'' \<and>
    mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>) }"

lemma critical_pairsI:
  assumes "(l, r) \<in> P" and "(l', r') \<in> R" and "l = C\<langle>l''\<rangle>"
    and "is_Fun l''" and "mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>)" and "s = r \<cdot> \<sigma>"
    and "t =  (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>" and "b = (C = \<box>)"
  shows "(b, s, t) \<in> critical_pairs P R"
  using assms unfolding critical_pairs_def by blast

lemma critical_pairs_main:
  fixes R :: "('f, string) trs"
  assumes st1: "(s, t1) \<in> rstep R" and st2: "(s, t2) \<in> rstep R"
  shows "(t1, t2) \<in> join (rstep R) \<or>
    (\<exists> C b l r \<sigma>. t1 = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t2 = C\<langle>r \<cdot> \<sigma>\<rangle> \<and>
    ((b, l, r) \<in> critical_pairs R R \<or> (b, r, l) \<in> critical_pairs R R))"
proof -
  let ?R = "rstep R"
  let ?CP = "critical_pairs R R"
  from st1 obtain C1 l1 r1 \<sigma>1 where lr1: "(l1, r1) \<in> R" and s1: "s = C1\<langle>l1 \<cdot> \<sigma>1\<rangle>" and t1: "t1 = C1\<langle>r1 \<cdot> \<sigma>1\<rangle>" by auto
  from st2 obtain C2 l2 r2 \<sigma>2 where lr2: "(l2, r2) \<in> R" and s2: "s = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>" and t2: "t2 = C2\<langle>r2 \<cdot> \<sigma>2\<rangle>" by auto
  from s1 s2 have id: "C1\<langle>l1 \<cdot> \<sigma>1\<rangle> = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>" by simp
  let ?q = "\<lambda> s t. (s, t) \<in> join ?R \<or> (\<exists> C b l r \<sigma>. s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> ((b, l, r) \<in> ?CP \<or> (b, r, l) \<in> ?CP))"
  let ?p = "\<lambda> (C1, C2). \<forall> l1 r1 l2 r2 \<sigma>1 \<sigma>2. (l1, r1) \<in> R \<longrightarrow> (l2, r2) \<in> R \<longrightarrow> C1\<langle>l1 \<cdot> \<sigma>1\<rangle> = C2\<langle>l2 \<cdot> \<sigma>2\<rangle> \<longrightarrow> ?q (C1\<langle>r1 \<cdot> \<sigma>1\<rangle>) (C2\<langle>r2 \<cdot> \<sigma>2\<rangle>)"
  {
    fix C12
    let ?m = "[\<lambda> (C1, C2). size C1 + size C2, \<lambda> (C1, C2). size C1]"
    have "?p C12" 
    proof (induct rule: wf_induct [OF wf_measures [of ?m], of ?p])
      case (1 C12)
      obtain C1 C2 where C12: "C12 = (C1, C2)" by force
      show "?p C12" unfolding C12 split
      proof (intro allI impI)
        fix l1 r1 l2 r2 \<sigma>1 \<sigma>2
        assume lr1: "(l1, r1) \<in> R" and lr2: "(l2, r2) \<in> R" and id: "C1\<langle>l1 \<cdot> \<sigma>1\<rangle> = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>"
        show "?q (C1\<langle>r1 \<cdot> \<sigma>1\<rangle>) (C2\<langle>r2 \<cdot> \<sigma>2\<rangle>)"
        proof (cases C1)
          case Hole note C1 = this
          with id have id: "l1 \<cdot> \<sigma>1 = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>" by simp
          let ?p = "hole_pos C2"
          show ?thesis
          proof (cases "?p \<in> poss l1 \<and> is_Fun (l1 |_ ?p)")
            case True
            hence p: "?p \<in> poss l1" by auto            
            from ctxt_supt_id [OF p] obtain D where Dl1: "D\<langle>l1 |_ ?p\<rangle> = l1" 
              and D: "D = ctxt_of_pos_term (hole_pos C2) l1" by blast
            from arg_cong [OF Dl1, of "\<lambda> t. t \<cdot> \<sigma>1"]
            have "(D \<cdot>c \<sigma>1)\<langle>(l1 |_ ?p) \<cdot> \<sigma>1\<rangle> = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>" unfolding id by simp
            from arg_cong [OF this, of "\<lambda> t. t |_ ?p"]
            have "l2 \<cdot> \<sigma>2 = (D \<cdot>c \<sigma>1)\<langle>(l1 |_ ?p) \<cdot> \<sigma>1\<rangle> |_ ?p" by simp
            also have "... = (D \<cdot>c \<sigma>1)\<langle>(l1 |_ ?p) \<cdot> \<sigma>1\<rangle> |_ (hole_pos (D \<cdot>c \<sigma>1))"
              using hole_pos_ctxt_of_pos_term [OF p] unfolding D by simp
            also have "... = (l1 |_ ?p) \<cdot> \<sigma>1" by (rule subt_at_hole_pos)
            finally have ident: "l2 \<cdot> \<sigma>2 = l1 |_ ?p \<cdot> \<sigma>1" by auto
            from mgu_var_disjoint_string_complete [OF ident [symmetric]]
            obtain \<mu>1 \<mu>2 \<rho> where mgu: "mgu_var_disjoint_string (l1 |_ ?p) l2 = Some (\<mu>1, \<mu>2)" and
              \<mu>1: "\<sigma>1 = \<mu>1 \<circ>s \<rho>"
              and \<mu>2: "\<sigma>2 = \<mu>2 \<circ>s \<rho>"
              and ident': "l1 |_ ?p \<cdot> \<mu>1 = l2 \<cdot> \<mu>2" by blast
            have in_cp: "(D = \<box>, r1 \<cdot> \<mu>1, (D \<cdot>c \<mu>1)\<langle>r2 \<cdot> \<mu>2\<rangle>) \<in> critical_pairs R R" 
              by (rule critical_pairsI [OF lr1 lr2 Dl1 [symmetric] True [THEN conjunct2] mgu], auto)
            from hole_pos_ctxt_of_pos_term [OF p] D have pD: "?p = hole_pos D" by simp
            from id have C2: "C2 = ctxt_of_pos_term ?p (l1 \<cdot> \<sigma>1)" by simp
            have "C2\<langle>r2 \<cdot> \<sigma>2\<rangle> = (ctxt_of_pos_term ?p (l1 \<cdot> \<sigma>1))\<langle>r2 \<cdot> \<sigma>2\<rangle>" using C2 by simp
            also have "... = (ctxt_of_pos_term ?p l1 \<cdot>c \<sigma>1)\<langle>r2 \<cdot> \<sigma>2\<rangle>" unfolding ctxt_of_pos_term_subst [OF p] ..
            also have "... = (D \<cdot>c \<sigma>1)\<langle>r2 \<cdot> \<sigma>2\<rangle>" unfolding D ..
            finally have C2r\<sigma>: "C2\<langle>r2 \<cdot> \<sigma>2\<rangle> = (D \<cdot>c \<sigma>1)\<langle>r2 \<cdot> \<sigma>2\<rangle>" .              
            from C1 have C1r\<sigma>: "C1\<langle>r1 \<cdot> \<sigma>1\<rangle> = r1 \<cdot> \<sigma>1" by simp
            show ?thesis unfolding C2r\<sigma> C1r\<sigma> unfolding \<mu>1 \<mu>2
            proof (rule disjI2, intro exI, intro conjI)
              show "r1 \<cdot> \<mu>1 \<circ>s \<rho> = \<box>\<langle>r1 \<cdot> \<mu>1 \<cdot> \<rho>\<rangle>" by simp
            qed (insert in_cp, auto)
          next
            case False
            from pos_into_subst [OF id _ False]
            obtain q q' x where p: "?p = q <#> q'" and q: "q \<in> poss l1" and l1q: "l1 |_ q = Var x" by auto
            have "l2 \<cdot> \<sigma>2 = C2\<langle>l2 \<cdot> \<sigma>2\<rangle> |_ q <#> q'" unfolding p [symmetric] by simp
            also have "... = l1 \<cdot> \<sigma>1 |_ q <#> q'" unfolding id ..
            also have "... = l1 |_ q \<cdot> \<sigma>1 |_ q'" using q by simp
            also have "... = \<sigma>1 x |_ q'" unfolding l1q by simp
            finally have l2x: "l2 \<cdot> \<sigma>2 = \<sigma>1 x |_ q'" by simp
            have pp: "?p \<in> poss (l1 \<cdot> \<sigma>1)" unfolding id by simp
            hence "q <#> q' \<in> poss (l1 \<cdot> \<sigma>1)" unfolding p .
            hence "q' \<in> poss (l1 \<cdot> \<sigma>1 |_ q)" unfolding poss_append_poss ..
            with q have "q' \<in> poss (l1 |_ q \<cdot> \<sigma>1)" by auto
            hence q'x: "q' \<in> poss (\<sigma>1 x)" unfolding l1q by simp
            from ctxt_supt_id [OF q'x] obtain E where \<sigma>1x: "E\<langle>l2 \<cdot> \<sigma>2\<rangle> = \<sigma>1 x"
              and E: "E = ctxt_of_pos_term q' (\<sigma>1 x)"
              unfolding l2x by blast
            let ?e = "E\<langle>r2 \<cdot> \<sigma>2\<rangle>"
            from hole_pos_ctxt_of_pos_term [OF q'x] E have q': "q' = hole_pos E" by simp
            from \<sigma>1x have \<sigma>1x': "\<sigma>1 x = E\<langle>l2 \<cdot> \<sigma>2\<rangle>" by simp
            let ?\<sigma> = "\<lambda> y. if y = x then ?e else \<sigma>1 y"
            have one: "(r1 \<cdot> \<sigma>1, r1 \<cdot> ?\<sigma>) \<in> (rstep R)^*" 
            proof (rule subst_qrsteps_imp_qrsteps [where Q = "{}", unfolded qrstep_rstep_conv])
              fix y
              show "(\<sigma>1 y, ?\<sigma> y) \<in> (rstep R)^*"
              proof (cases "y = x")
                case False thus ?thesis by simp
              next
                case True
                show ?thesis unfolding True \<sigma>1x' using lr2 by auto
              qed
            qed            
            show ?thesis 
            proof (rule disjI1, rule, unfold C1 ctxt_apply.simps, rule one)
              show "(C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, r1 \<cdot> ?\<sigma>) \<in> ?R^*" 
              proof (rule rtrancl_trans)
                show "(l1 \<cdot> ?\<sigma>, r1 \<cdot> ?\<sigma>) \<in> ?R^*" using lr1 by auto
              next
                from q have ql1:  "q \<in> poss (l1 \<cdot> \<sigma>1)" by simp                
                have "C2\<langle>r2 \<cdot> \<sigma>2\<rangle> = replace_at (C2\<langle>l2 \<cdot> \<sigma>2\<rangle>) ?p (r2 \<cdot> \<sigma>2)" by simp
                also have "... = replace_at (l1 \<cdot> \<sigma>1) ?p (r2 \<cdot> \<sigma>2)" unfolding id ..
                also have "... = replace_at (l1 \<cdot> \<sigma>1) q ?e"
                proof -
                  have "E = ctxt_of_pos_term q' (l1 \<cdot> \<sigma>1 |_ q)" 
                    unfolding subt_at_subst [OF q] l1q E by simp
                  thus ?thesis
                    unfolding p                  
                    unfolding ctxt_of_pos_term_append [OF ql1]
                    by simp
                qed
                finally have CR2: "C2\<langle>r2 \<cdot> \<sigma>2\<rangle> = replace_at (l1 \<cdot> \<sigma>1) q ?e" .
                from q l1q have "(replace_at (l1 \<cdot> \<sigma>1) q ?e, l1 \<cdot> ?\<sigma>) \<in> ?R^*"
                proof (induct l1 arbitrary: q)
                  case (Var y)
                  thus ?case by simp
                next
                  case (Fun f ls)
                  from Fun(2, 3) obtain i p where q: "q = i <# p" and i: "i < length ls" and p: "p \<in> poss (ls ! i)" and px: "ls ! i |_ p = Var x" by (cases q, auto)
                  from i have "ls ! i \<in> set ls" by auto
                  from Fun(1) [OF this p px] have rec: "(replace_at (ls ! i \<cdot> \<sigma>1) p ?e, ls ! i \<cdot> ?\<sigma>) \<in> ?R^*" .
                  let ?ls\<sigma> = "map (\<lambda> t. t \<cdot> \<sigma>1) ls"
                  let ?ls\<sigma>' = "map (\<lambda> t. t \<cdot> ?\<sigma>) ls"
                  have id: "replace_at (Fun f ls \<cdot> \<sigma>1) q ?e = Fun f (take i ?ls\<sigma> @ replace_at (ls ! i \<cdot> \<sigma>1) p ?e # drop (Suc i) ?ls\<sigma>)" (is "_ = Fun f ?r")
                    unfolding q using i by simp
                  show ?case unfolding id unfolding subst_apply_term.simps
                  proof (rule all_ctxt_closedE [of UNIV])
                    fix j
                    assume j: "j < length ?r"
                    show "(?r ! j, ?ls\<sigma>' ! j) \<in> ?R^*"
                    proof (cases "j = i")
                      case True
                      show ?thesis using i True using rec by (auto simp: nth_append)
                    next
                      case False
                      have "?r ! j = ?ls\<sigma> ! j" 
                        by (rule nth_append_take_drop_is_nth_conv, insert False i j, auto)
                      also have "... = ls ! j \<cdot> \<sigma>1" using j i by auto
                      finally have idr: "?r ! j = ls ! j \<cdot> \<sigma>1" .
                      from j i have idl: "?ls\<sigma>' ! j = ls ! j \<cdot> ?\<sigma>" by auto
                      show ?thesis unfolding idr idl
                      proof (rule subst_qrsteps_imp_qrsteps [where Q = "{}", unfolded qrstep_rstep_conv])
                        fix y
                        show "(\<sigma>1 y, ?\<sigma> y) \<in> ?R^*"
                        proof (cases "y = x")
                          case False thus ?thesis by simp
                        next
                          case True thus ?thesis using \<sigma>1x' lr2 by auto
                        qed
                      qed
                    qed
                  qed (insert i, auto)
                qed
                thus "(C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, l1 \<cdot> ?\<sigma>) \<in> ?R^*" unfolding CR2 .
              qed
            qed
          qed
        next
          case (More f1 bef1 D1 aft1) note C1 = this
          show ?thesis
          proof (cases C2)
            case Hole note C2 = this
            from id have id: "C2\<langle>l2 \<cdot> \<sigma>2\<rangle> = C1\<langle>l1 \<cdot> \<sigma>1\<rangle>" by simp
            from C1 C2 have "((C2, C1), C12) \<in> measures ?m" unfolding C12 by auto
            from 1 [rule_format, OF this, unfolded split, rule_format, OF lr2 lr1 id]
            show ?thesis 
            proof 
              assume "(C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, C1\<langle>r1 \<cdot> \<sigma>1\<rangle>) \<in> join ?R"
              thus ?thesis by auto            
            qed blast
          next
            case (More f2 bef2 D2 aft2) note C2 = this
            let ?n1 = "length bef1"
            let ?n2 = "length bef2"
            note id = id [unfolded C1 C2]
            from id have f: "f1 = f2" by simp
            show ?thesis
            proof (cases "?n1 = ?n2")
              case True
              with id have idb: "bef1 = bef2" and ida: "aft1 = aft2" 
                and idD: "D1\<langle>l1 \<cdot> \<sigma>1\<rangle> = D2\<langle>l2 \<cdot> \<sigma>2\<rangle>" by auto
              let ?C = "More f2 bef2 \<box> aft2"
              have id1: "C1 = ?C \<circ>c D1" unfolding C1 f ida idb by simp
              have id2: "C2 = ?C \<circ>c D2" unfolding C2 by simp
              have "((D1, D2), C12) \<in> measures ?m" unfolding C12 C1 C2
                by auto
              from 1 [rule_format, OF this, unfolded split, rule_format, 
                OF lr1 lr2 idD] have "?q (D1\<langle>r1 \<cdot> \<sigma>1\<rangle>) (D2\<langle>r2 \<cdot> \<sigma>2\<rangle>)" by auto
              thus ?thesis
              proof 
                assume "(D1\<langle>r1 \<cdot> \<sigma>1\<rangle>, D2\<langle>r2 \<cdot> \<sigma>2\<rangle>) \<in> join ?R" 
                then obtain s' where seq1: "(D1\<langle>r1 \<cdot> \<sigma>1\<rangle>, s') \<in> ?R^*" 
                and seq2: "(D2\<langle>r2 \<cdot> \<sigma>2\<rangle>, s') \<in> ?R^*" by auto
                from rsteps_closed_ctxt [OF seq1, of ?C]
                have seq1: "(C1\<langle>r1 \<cdot> \<sigma>1\<rangle>, ?C\<langle>s'\<rangle>) \<in> ?R^*" using id1 by auto
                from rsteps_closed_ctxt [OF seq2, of ?C]              
                have seq2: "(C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, ?C\<langle>s'\<rangle>) \<in> ?R^*" using id2 by auto
                from seq1 seq2 show ?thesis by auto
              next
                assume "\<exists> C b l r \<sigma>. D1\<langle>r1 \<cdot> \<sigma>1\<rangle> = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> D2\<langle>r2 \<cdot> \<sigma>2\<rangle> = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> ((b, l, r) \<in> ?CP \<or> (b, r, l) \<in> ?CP)"
                then obtain C b l r \<sigma> where idD: "D1\<langle>r1 \<cdot> \<sigma>1\<rangle> = C\<langle>l \<cdot> \<sigma>\<rangle>" "D2\<langle>r2 \<cdot> \<sigma>2\<rangle> = C\<langle>r \<cdot> \<sigma>\<rangle>" and mem: "((b, l, r) \<in> ?CP \<or> (b, r, l) \<in> ?CP)" by blast
                show ?thesis 
                  by (rule disjI2, unfold id1 id2, rule exI [of _ "?C \<circ>c C"], intro exI, rule conjI [OF _ conjI [OF _ mem]], insert idD, auto)
              qed
            next
              case False
              let ?p1 = "?n1 <# hole_pos D1"
              let ?p2 = "?n2 <# hole_pos D2"
              have l2: "C1\<langle>l1 \<cdot> \<sigma>1\<rangle> |_ ?p2 = l2 \<cdot> \<sigma>2" unfolding C1 id by simp   
              have p12: "?p1  \<bottom> ?p2" using False by simp
              have p1: "?p1 \<in> poss (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>)" unfolding C1 by simp
              have p2: "?p2 \<in> poss (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>)" unfolding C1 unfolding id by simp
              let ?one = "replace_at (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>) ?p1 (r1 \<cdot> \<sigma>1)"
              have one: "C1\<langle>r1 \<cdot> \<sigma>1\<rangle> = ?one" unfolding C1 by simp
              from parallel_qrstep [OF p12 p1 p2 l2 _ lr2, of "{}" True "r1 \<cdot> \<sigma>1"] 
              have "(?one, replace_at ?one ?p2 (r2 \<cdot> \<sigma>2)) \<in> rstep R" by auto
              hence one: "(C1\<langle>r1 \<cdot> \<sigma>1\<rangle>, replace_at ?one ?p2 (r2 \<cdot> \<sigma>2)) \<in> (rstep R)^*" unfolding one by simp
              have l1: "C2\<langle>l2 \<cdot> \<sigma>2\<rangle> |_ ?p1 = l1 \<cdot> \<sigma>1" unfolding C2 id [symmetric] by simp
              have p21: "?p2  \<bottom> ?p1" using False by simp
              have p1': "?p1 \<in> poss (C2\<langle>l2 \<cdot> \<sigma>2\<rangle>)" unfolding C2 id [symmetric] by simp
              have p2': "?p2 \<in> poss (C2\<langle>l2 \<cdot> \<sigma>2\<rangle>)" unfolding C2 by simp
              let ?two = "replace_at (C2\<langle>l2 \<cdot> \<sigma>2\<rangle>) ?p2 (r2 \<cdot> \<sigma>2)"
              have two: "C2\<langle>r2 \<cdot> \<sigma>2\<rangle> = ?two" unfolding C2 by simp
              from parallel_qrstep [OF p21 p2' p1' l1 _ lr1, of "{}" True "r2 \<cdot> \<sigma>2"] 
              have "(?two, replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)) \<in> rstep R" by auto
              hence two: "(C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)) \<in> (rstep R)^*" unfolding two by simp
              have "replace_at ?one ?p2 (r2 \<cdot> \<sigma>2) = replace_at (replace_at (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>) ?p2 (r2 \<cdot> \<sigma>2)) ?p1 (r1 \<cdot> \<sigma>1)"
                by (rule parallel_replace_at [OF p12 p1 p2])
              also have "... = replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)" unfolding C1 C2 id ..
              finally have one_two: "replace_at ?one ?p2 (r2 \<cdot> \<sigma>2) = replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)" .
              show ?thesis
                by (rule disjI1, rule, rule one, unfold one_two, rule two)
            qed
          qed
        qed
      qed
    qed
  }
  from this [of "(C1, C2)", unfolded split, rule_format, OF lr1 lr2 id]
  show ?thesis unfolding t1 t2 by auto
qed

lemma critical_pairs:
  fixes R :: "('f, string) trs"
  assumes cp: "\<And> l r b. (b, l, r) \<in> critical_pairs R R \<Longrightarrow> l \<noteq> r \<Longrightarrow>
    \<exists> l' r' s. instance_rule (l, r) (l', r') \<and> (l', s) \<in> (rstep R)\<^sup>* \<and> (r', s) \<in> (rstep R)\<^sup>*"
  shows "WCR (rstep R)"
proof 
  let ?R = "rstep R"
  let ?CP = "critical_pairs R R"
  fix s t1 t2
  assume steps: "(s, t1) \<in> ?R" "(s, t2) \<in> ?R"
  let ?p = "\<lambda> s'. (t1, s') \<in> ?R^* \<and> (t2, s') \<in> ?R^*"
  from critical_pairs_main [OF steps] 
  have "\<exists> s'. ?p s'" 
  proof
    assume "\<exists> C b l r \<sigma>. t1 = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t2 = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> ((b, l, r) \<in> ?CP \<or> (b, r, l) \<in> ?CP)"
    then obtain C b l r \<sigma> where id: "t1 = C\<langle>l \<cdot> \<sigma>\<rangle>" "t2 = C\<langle>r \<cdot> \<sigma>\<rangle>" 
      and mem: "(b, l, r) \<in> ?CP \<or> (b, r, l) \<in> ?CP" by blast
    show ?thesis
    proof (cases "l = r")
      case True
      thus ?thesis unfolding id by auto
    next
      case False
      note sub_ctxt = rsteps_closed_ctxt [OF rsteps_closed_subst [OF rsteps_closed_subst]]
      from mem show ?thesis
      proof
        assume mem: "(b, l, r) \<in> ?CP"
        from cp [OF mem False] obtain l' r' s' \<tau> where id2: "l = l' \<cdot> \<tau>" "r = r' \<cdot> \<tau>" and steps: "(l', s') \<in> ?R^*" "(r', s') \<in> ?R^*" 
          unfolding instance_rule_def by auto
        show "\<exists> s'. ?p s'" unfolding id id2
          by (rule exI [of _ "C\<langle>s' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>"], rule conjI, rule sub_ctxt [OF steps(1)], rule sub_ctxt [OF steps(2)])
      next
        assume mem: "(b, r, l) \<in> ?CP"
        from cp [OF mem] False obtain l' r' s' \<tau> where id2: "r = l' \<cdot> \<tau>" "l = r' \<cdot> \<tau>" and steps: "(l', s') \<in> ?R^*" "(r', s') \<in> ?R^*" 
          unfolding instance_rule_def by auto
        show "\<exists> s'. ?p s'" unfolding id id2
          by (rule exI [of _ "C\<langle>s' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>"], rule conjI, rule sub_ctxt [OF steps(2)], rule sub_ctxt [OF steps(1)])
      qed
    qed
  qed auto
  thus "(t1, t2) \<in> join ?R" by auto
qed

lemma critical_pairs_fork:
  fixes R :: "('f, string) trs"
  assumes cp: "(b, l, r) \<in> critical_pairs R R"
  shows "(l, r) \<in> (rstep R)\<inverse> O rstep R"
proof -
  from cp [unfolded critical_pairs_def, simplified] 
  obtain \<sigma>1 \<sigma>2 l1 l2 r1 r2 C where id: "l = r1 \<cdot> \<sigma>1" "r = (C \<cdot>c \<sigma>1)\<langle>r2 \<cdot> \<sigma>2\<rangle>" 
    and r1: "(C\<langle>l1\<rangle>, r1) \<in> R" and r2: "(l2, r2) \<in> R" and mgu: "mgu_var_disjoint_string l1 l2 = Some (\<sigma>1, \<sigma>2)" by blast
  have one: "(l, C\<langle>l1\<rangle> \<cdot> \<sigma>1) \<in> (rstep R)^-1" unfolding reverseTrs id
    by (rule rstepI [of r1 "C\<langle>l1\<rangle>" _ _ \<box> \<sigma>1], insert r1, auto)
  from mgu_var_disjoint_generic_sound [OF mgu] have change: "C\<langle>l1\<rangle> \<cdot> \<sigma>1 = (C \<cdot>c \<sigma>1)\<langle>l2 \<cdot> \<sigma>2\<rangle>" by simp
  have "(C\<langle>l1\<rangle> \<cdot> \<sigma>1, r) \<in> rstep R" unfolding change id
    by (rule rstepI [OF r2, of _ _ \<sigma>2], auto)
  with one show ?thesis by auto
qed

(* in the following lemma infiniteness of strings is crucial, if one restricts to well-formed terms:
   Consider that we only have 5 variables and build the following TRS R = {
     f(g(x1, x2), g(x3, x4)) \<rightarrow> h(x1, h(x2, g(x3, x4)))
     f(g(g(x1, x2), g(x3, x4)), x5) \<rightarrow> h(x5, h(g(x1, x2), g(x3, x4)))
   the following rules are added to allow a join if one the arguments of
   the g's is a not a variable, 
     h(g(k(_, _), _), h(_, _)) \<rightarrow> ok for all k \<in> {f, g, h}
     h(g(_, k(_, _)), h(_, _)) \<rightarrow> ok for all k \<in> {f, g, h}
     h(_, h(g(k(_, _), _), _)) \<rightarrow> ok for all k \<in> {f, g, h}
     h(_, h(g(_, k(_, _)), _)) \<rightarrow> ok for all k \<in> {f, g, h}
     h(_, h(_, g(k(_, _), _))) \<rightarrow> ok for all k \<in> {f, g, h}
     h(_, h(_, g(_, k(_, _)))) \<rightarrow> ok for all k \<in> {f, g, h}
     h(g(ok, _), h(_, _)) \<rightarrow> ok 
     h(g(_, ok), h(_, _)) \<rightarrow> ok
     h(_, h(g(ok, _), _)) \<rightarrow> ok
     h(_, h(g(_, ok), _)) \<rightarrow> ok
     h(_, h(_, g(ok, _))) \<rightarrow> ok
     h(_, h(_, g(_, ok))) \<rightarrow> ok
   and we allow a join if two of the six arguments are identical
     h(g(x1, x2), h(g(x3, x4), g(x5, x))) \<rightarrow> ok for all x \<in> {x1, .., x5}
     h(g(x1, x2), h(g(x3, x4), g(x, x5))) \<rightarrow> ok for all x \<in> {x1, .., x5}
     h(g(x1, x2), h(g(x3, x), g(x4, x5))) \<rightarrow> ok for all x \<in> {x1, .., x5}
     h(g(x1, x2), h(g(x, x3), g(x4, x5))) \<rightarrow> ok for all x \<in> {x1, .., x5}
     h(g(x1, x), h(g(x2, x3), g(x4, x5))) \<rightarrow> ok for all x \<in> {x1, .., x5}
     h(g(x, x1), h(g(x2, x3), g(x4, x5))) \<rightarrow> ok for all x \<in> {x1, .., x5}
   moreover, we add one rule to join some other critical pairs
     h(x1, ok) \<rightarrow> ok

  note that R is not WCR over the signature {f, g, h, ok} and variables {x1, .., x6} since in the critical pair
  h(g(x1, x2), h(g(x3, x4), g(x5, x6))) \<leftarrow> f(g(g(x1, x2), g(x3, x4)), g(x5, x6)) \<rightarrow> h(g(x5, x6), h(g(x1, x2), g(x3, x4)))
  both h-terms are normal forms.
  However, one can prove that R over the signature {f, g, h, ok} and variables {x1, .., x5} is WCR,
  since the critical pair above cannot be build using only 5 variables, and if one of the arguments is chosen
  as non-variable, or if two arguments are identical, then both terms can be reduced to ok.
  Moreover, all over overlaps can easily be reduced to ok, too.
*)
lemma critical_pairs_complete:
  fixes R :: "('f, string) trs"
  assumes cp: "(b, l, r) \<in> critical_pairs R R"
    and no_join: "(l, r) \<notin> (rstep R)\<^sup>\<down>"
  shows "\<not> WCR (rstep R)"
proof 
  from critical_pairs_fork [OF cp] obtain u where ul: "(u, l) \<in> rstep R" and ur: "(u, r) \<in> rstep R" by force  
  assume wcr: "WCR (rstep R)"
  with ul ur no_join show False unfolding WCR_on_def by auto
qed

lemma critical_pair_lemma:
  fixes R :: "('f, string) trs"
  shows "WCR (rstep R) \<longleftrightarrow>
    (\<forall> (b, s, t) \<in> critical_pairs R R. (s, t) \<in> (rstep R)\<^sup>\<down>)"
  (is "?l = ?r")
proof 
  assume ?l
  with critical_pairs_complete [where R = R] show ?r by auto
next
  assume ?r
  show ?l
  proof (rule critical_pairs)
    fix b s t 
    assume "(b, s, t) \<in> critical_pairs R R"
    with `?r` have "(s, t) \<in> join (rstep R)" by auto
    then obtain u where s: "(s, u) \<in> (rstep R)^*"
      and t: "(t, u) \<in> (rstep R)^*" by auto
    show "\<exists> s' t' u. instance_rule (s, t) (s', t') \<and> (s', u) \<in> (rstep R)^* \<and> (t', u) \<in> (rstep R)^*"
    proof (intro exI conjI)
      show "instance_rule (s, t) (s, t)" by simp
    qed (insert s t, auto)
  qed
qed

(* note that the lhss condition is necessary: 
   consider R = {x \<rightarrow> f(a), x \<rightarrow> g(a)} and Q = {x}
   then there are no critical pairs and NF Q \<subseteq> NF R
   however, f(a) \<leftarrow> x \<rightarrow> g(a) and f(a) and g(a) are not joinable *)
lemma critical_pairs_innermost_weak_diamond:
  fixes R :: "('f, string) trs"
  assumes cp: "\<And> l r. (True, l, r) \<in> critical_pairs R R \<Longrightarrow> l = r"
    and NF_Q_R: "NF_terms Q \<subseteq> NF_trs R"
    and lhss: "\<And> l r. (l, r) \<in> R \<Longrightarrow> is_Fun l"
  shows "w\<diamond> (qrstep nfs Q R)"
unfolding weak_diamond_def
proof 
  fix t1 t2 :: "('f, string) term"
  let ?R = "qrstep nfs Q R"
  assume "(t1, t2) \<in> ?R^-1 O ?R - Id"
  then obtain s where st1: "(s, t1) \<in> ?R" and st2: "(s, t2) \<in> ?R" and t12: "t1 \<noteq> t2" by auto
  let ?Q = "NF_terms Q"
  from st1 obtain C1 l1 r1 \<sigma>1 where lr1: "(l1, r1) \<in> R" and s1: "s = C1\<langle>l1 \<cdot> \<sigma>1\<rangle>" and t1: "t1 = C1\<langle>r1 \<cdot> \<sigma>1\<rangle>" 
    and NF1: "\<forall> u \<lhd> l1 \<cdot> \<sigma>1. u \<in> ?Q" and nfs1: "NF_subst nfs (l1, r1) \<sigma>1 Q" by auto
  from st2 obtain C2 l2 r2 \<sigma>2 where lr2: "(l2, r2) \<in> R" and s2: "s = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>" and t2: "t2 = C2\<langle>r2 \<cdot> \<sigma>2\<rangle>" 
    and NF2: "\<forall> u \<lhd> l2 \<cdot> \<sigma>2. u \<in> ?Q" and nfs2: "NF_subst nfs (l2, r2) \<sigma>2 Q" by auto
  from s1 s2 have id: "C1\<langle>l1 \<cdot> \<sigma>1\<rangle> = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>" by simp
  let ?p = "\<lambda> (C1, C2). C1\<langle>l1 \<cdot> \<sigma>1\<rangle> = C2\<langle>l2 \<cdot> \<sigma>2\<rangle> \<longrightarrow> (\<exists> s'. (C1\<langle>r1 \<cdot> \<sigma>1\<rangle>, s') \<in> ?R \<and> (C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, s') \<in> ?R \<or> C1\<langle>r1 \<cdot> \<sigma>1\<rangle> = C2\<langle>r2 \<cdot> \<sigma>2\<rangle>)"
  {
    fix C12
    let ?m = "\<lambda> (C1, C2). size C1 + size C2"
    have "?p C12" 
    proof (induct rule: wf_induct [OF wf_measure [of ?m], of ?p])
      case (1 C12)
      obtain C1 C2 where C12: "C12 = (C1, C2)" by force
      show "?p C12" unfolding C12 split
      proof (intro impI)
        assume id: "C1\<langle>l1 \<cdot> \<sigma>1\<rangle> = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>"
        show "\<exists> s'. ( (C1\<langle>r1 \<cdot> \<sigma>1\<rangle>, s') \<in> ?R \<and> (C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, s') \<in> ?R \<or> C1\<langle>r1 \<cdot> \<sigma>1\<rangle> = C2\<langle>r2 \<cdot> \<sigma>2\<rangle>)"
        proof (cases C1)
          case Hole note C1 = this
          with id have id: "l1 \<cdot> \<sigma>1 = C2\<langle>l2 \<cdot> \<sigma>2\<rangle>" by simp
          have C2: "C2 = \<box>"
          proof (rule ccontr)
            assume "C2 \<noteq> \<box>"
            with id have "l2 \<cdot> \<sigma>2 \<lhd> l1 \<cdot> \<sigma>1" by auto
            with NF1 NF_Q_R have "l2 \<cdot> \<sigma>2 \<in> NF_trs R" by auto
            with lr2 show False by auto
          qed
          with id have ident: "l1 \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2"  by simp
          from lhss [OF lr1] have nvar: "is_Fun l1" .
          from mgu_var_disjoint_string_complete [OF ident]
          obtain \<mu>1 \<mu>2 \<rho> where mgu: "mgu_var_disjoint_string l1 l2 = Some (\<mu>1, \<mu>2)" and
            \<mu>1: "\<sigma>1 = \<mu>1 \<circ>s \<rho>"
            and \<mu>2: "\<sigma>2 = \<mu>2 \<circ>s \<rho>" 
            by blast
          have in_cp: "(True, r1 \<cdot> \<mu>1, r2 \<cdot> \<mu>2) \<in> critical_pairs R R" 
            by (rule critical_pairsI [OF lr1 lr2 _ nvar mgu, of \<box>], auto)
          from C2 have C2r\<sigma>: "C2\<langle>r2 \<cdot> \<sigma>2\<rangle> = r2 \<cdot> \<sigma>2" by simp
          from C1 have C1r\<sigma>: "C1\<langle>r1 \<cdot> \<sigma>1\<rangle> = r1 \<cdot> \<sigma>1" by simp
          from cp [OF in_cp, unfolded instance_rule_def] have id: "r1 \<cdot> \<mu>1 = r2 \<cdot> \<mu>2" .
          from C1r\<sigma> have "C1\<langle>r1 \<cdot> \<sigma>1\<rangle> = r2 \<cdot> \<sigma>2" unfolding \<mu>1 using id \<mu>2 by simp
          also have "... = C2\<langle>r2 \<cdot> \<sigma>2\<rangle>" unfolding C2r\<sigma> ..
          finally show ?thesis by simp
        next
          case (More f1 bef1 D1 aft1) note C1 = this
          show ?thesis
          proof (cases C2)
            case Hole 
            with id have "l2 \<cdot> \<sigma>2 = C1\<langle>l1 \<cdot> \<sigma>1\<rangle>" by auto
            with C1 have "l1 \<cdot> \<sigma>1 \<lhd> l2 \<cdot> \<sigma>2" by auto
            with NF2 NF_Q_R have "l1 \<cdot> \<sigma>1 \<in> NF_trs R" by auto
            with lr1 have False by auto
            thus ?thesis ..
          next
            case (More f2 bef2 D2 aft2) note C2 = this
            let ?n1 = "length bef1"
            let ?n2 = "length bef2"
            note id = id [unfolded C1 C2]
            from id have f: "f1 = f2" by simp
            show ?thesis
            proof (cases "?n1 = ?n2")
              case True
              with id have idb: "bef1 = bef2" and ida: "aft1 = aft2" 
                and idD: "D1\<langle>l1 \<cdot> \<sigma>1\<rangle> = D2\<langle>l2 \<cdot> \<sigma>2\<rangle>" by auto
              have "((D1, D2), C12) \<in> measure ?m" unfolding C12 C1 C2
                by auto
              from 1 [rule_format, OF this, unfolded split, rule_format, 
                OF idD] obtain s' 
                where disj: "(D1\<langle>r1 \<cdot> \<sigma>1\<rangle>, s') \<in> ?R \<and> (D2\<langle>r2 \<cdot> \<sigma>2\<rangle>, s') \<in> ?R \<or> D1\<langle>r1 \<cdot> \<sigma>1\<rangle> = D2\<langle>r2 \<cdot> \<sigma>2\<rangle>" (is "?seq1 \<and> ?seq2 \<or> ?id") by auto 
              let ?C = "More f2 bef2 \<box> aft2"
              have id1: "C1 = ?C \<circ>c D1" unfolding C1 f ida idb by simp
              have id2: "C2 = ?C \<circ>c D2" unfolding C2 by simp
              from disj show ?thesis
              proof
                assume "?seq1 \<and> ?seq2" 
                hence seq1: "?seq1" and seq2: "?seq2" by auto
                from qrstep_closed_ctxt [OF seq1, of ?C]
                have seq1: "(C1\<langle>r1 \<cdot> \<sigma>1\<rangle>, ?C\<langle>s'\<rangle>) \<in> ?R" using id1 by auto
                from qrstep_closed_ctxt [OF seq2, of ?C]              
                have seq2: "(C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, ?C\<langle>s'\<rangle>) \<in> ?R" using id2 by auto
                from seq1 seq2 show ?thesis by auto
              next
                assume ?id
                thus ?thesis unfolding id1 id2 by simp
              qed
            next
              case False
              let ?p1 = "?n1 <# hole_pos D1"
              let ?p2 = "?n2 <# hole_pos D2"
              have l2: "C1\<langle>l1 \<cdot> \<sigma>1\<rangle> |_ ?p2 = l2 \<cdot> \<sigma>2" unfolding C1 id by simp   
              have p12: "?p1  \<bottom> ?p2" using False by simp
              have p1: "?p1 \<in> poss (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>)" unfolding C1 by simp
              have p2: "?p2 \<in> poss (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>)" unfolding C1 unfolding id by simp
              let ?one = "replace_at (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>) ?p1 (r1 \<cdot> \<sigma>1)"
              have one: "C1\<langle>r1 \<cdot> \<sigma>1\<rangle> = ?one" unfolding C1 by simp
              from parallel_qrstep [OF p12 p1 p2 l2 NF2 lr2 nfs2] 
              have "(?one, replace_at ?one ?p2 (r2 \<cdot> \<sigma>2)) \<in> qrstep nfs Q R" .
              hence one: "(C1\<langle>r1 \<cdot> \<sigma>1\<rangle>, replace_at ?one ?p2 (r2 \<cdot> \<sigma>2)) \<in> qrstep nfs Q R" unfolding one by simp
              have l1: "C2\<langle>l2 \<cdot> \<sigma>2\<rangle> |_ ?p1 = l1 \<cdot> \<sigma>1" unfolding C2 id [symmetric] by simp
              have p21: "?p2  \<bottom> ?p1" using False by simp
              have p1': "?p1 \<in> poss (C2\<langle>l2 \<cdot> \<sigma>2\<rangle>)" unfolding C2 id [symmetric] by simp
              have p2': "?p2 \<in> poss (C2\<langle>l2 \<cdot> \<sigma>2\<rangle>)" unfolding C2 by simp
              let ?two = "replace_at (C2\<langle>l2 \<cdot> \<sigma>2\<rangle>) ?p2 (r2 \<cdot> \<sigma>2)"
              have two: "C2\<langle>r2 \<cdot> \<sigma>2\<rangle> = ?two" unfolding C2 by simp
              from parallel_qrstep [OF p21 p2' p1' l1 NF1 lr1 nfs1] 
              have "(?two, replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)) \<in> qrstep nfs Q R" .
              hence two: "(C2\<langle>r2 \<cdot> \<sigma>2\<rangle>, replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)) \<in> qrstep nfs Q R" unfolding two by simp
              have "replace_at ?one ?p2 (r2 \<cdot> \<sigma>2) = replace_at (replace_at (C1\<langle>l1 \<cdot> \<sigma>1\<rangle>) ?p2 (r2 \<cdot> \<sigma>2)) ?p1 (r1 \<cdot> \<sigma>1)"
                by (rule parallel_replace_at [OF p12 p1 p2])
              also have "... = replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)" unfolding C1 C2 id ..
              finally have one_two: "replace_at ?one ?p2 (r2 \<cdot> \<sigma>2) = replace_at ?two ?p1 (r1 \<cdot> \<sigma>1)" .
              show ?thesis
                by (intro exI disjI1 conjI, rule one, unfold one_two, rule two)
            qed
          qed
        qed
      qed
    qed
  }
  from this [of "(C1, C2)", unfolded split, rule_format, OF id]
  show "(t1, t2) \<in> ?R O ?R^-1" using t12 unfolding t1 t2 by auto
qed

lemma critical_pairs_innermost:
  assumes "\<And> l r. (True, l, r) \<in> critical_pairs R R \<Longrightarrow> l = r"
    and "NF_terms Q \<subseteq> NF_trs R"
    and "\<And> l r. (l, r) \<in> R \<Longrightarrow> is_Fun l"
  shows "CR (qrstep nfs Q R)"
  by (rule weak_diamond_imp_CR [OF critical_pairs_innermost_weak_diamond [OF assms]])

lemma critical_pairs_exI:
  fixes \<sigma> :: "('f, string) subst"
  assumes P: "(l, r) \<in> P" and R: "(l', r') \<in> R" and l: "l = C\<langle>l''\<rangle>"
    and l'': "is_Fun l''" and unif: "l'' \<cdot> \<sigma> = l' \<cdot> \<tau>"
    and b: "b = (C = \<box>)"
  shows "\<exists> s t. (b, s, t) \<in> critical_pairs P R"
proof - 
  from mgu_var_disjoint_string_complete [OF unif]
  obtain \<mu>1 \<mu>2  where mgu: "mgu_var_disjoint_string l'' l' = Some (\<mu>1, \<mu>2)" by blast
  show ?thesis
    by (intro exI, rule critical_pairsI [OF P R l l'' mgu refl refl b])
qed
  
end

