(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multiset2
imports
  "~~/src/HOL/Library/Multiset"
  Relations 
begin

abbreviation(input) mset_sum where "mset_sum \<equiv> Multiset.fold (op +) (0 :: 'a :: comm_monoid_add)"

lemma comp_fun_commute_plus[simp]: "comp_fun_commute (op + :: (('a :: comm_monoid_add) \<Rightarrow> 'a \<Rightarrow> 'a))"
  by (default, auto simp: ac_simps)

lemma [simp]: "mset_sum (M + N) = mset_sum M + mset_sum N"
  unfolding comp_fun_commute.fold_mset_union [OF comp_fun_commute_plus]
  unfolding comp_fun_commute.fold_mset_fun_left_comm [OF comp_fun_commute_plus]
  by simp

lemma [simp]: "mset_sum {# x #} = x"
proof -
  have "mset_sum ({#x#}) = mset_sum ({#} + {#x#})" by simp
  also have "... = x" 
    unfolding comp_fun_commute.fold_mset_insert[OF comp_fun_commute_plus] by simp
  finally show ?thesis .
qed

lemma union_commutes:
  "M + {#x#} + N = M + N + {#x#}"
  "M + multiset_of xs + N = M + N + multiset_of xs"
  by (auto simp: ac_simps)

lemma in_multiset_of: assumes "x \<in># multiset_of xs"
  shows "\<exists> bef aft. xs = bef @ x # aft"
proof -
  from assms[unfolded in_multiset_in_set set_conv_nth]
  obtain i where i: "i < length xs" and x: "xs ! i = x" by auto
  from id_take_nth_drop[OF i, unfolded x]
  show ?thesis by auto
qed

lemma in_multiset_of_idx :
  assumes "a \<in># multiset_of as"
  shows "\<exists>i. i < length as \<and> a = as ! i"
proof-
 from assms and in_multiset_of[of as a] obtain bef aft where
   as_part: "as = bef @ a # aft" by auto
 with nth_append_length have a_eq: "a = as ! (length bef)" by auto
 from as_part and length_append have i_bd: "(length bef) < length as" by auto
 from a_eq and i_bd show ?thesis by blast
qed

lemma in_set_idx:
  assumes "a \<in> set as"
  shows "\<exists>i. i < length as \<and> a = as ! i"
using assms and in_multiset_in_set[of as a] and in_multiset_of_idx[of as a]
by simp

lemma mem_multiset_diff_compat:
  assumes "x \<in># A"
  and "x \<noteq> y"
  shows "x \<in># (A - {#y#})"
using assms by (induct A) auto

lemma in_multiset_of_eq: assumes "M + {#x#} = multiset_of ys"
  shows "\<exists> bef aft. ys = bef @ x # aft \<and> M = multiset_of (bef @ aft)"
proof -
  from assms[symmetric] have "x \<in># multiset_of ys" by auto
  from in_multiset_of[OF this] obtain bef aft where ys: "ys = bef @ x # aft"
    by auto
  have "M + {#x#} = multiset_of (bef @ aft) + {#x#}"
    unfolding assms ys multiset_eq_iff by auto
  hence "M = multiset_of (bef @ aft)" by auto
  from ys this show ?thesis by auto
qed

lemma multiset_of_concat: assumes "multiset_of xs = multiset_of ys"
  shows "multiset_of (concat xs) = multiset_of (concat ys)"
  using assms
proof (induct xs arbitrary: ys)
  case Nil thus ?case by auto
next
  case (Cons x xs)
  let ?m = "multiset_of"
  from in_multiset_of_eq[OF Cons(2)[simplified]]
  obtain bef aft where ys: "ys = bef @ x # aft" and xs: "?m xs = ?m (bef @ aft)"
    by auto
  from Cons(1)[OF xs] show ?case unfolding ys by (auto simp: multiset_eq_iff)
qed

lemma in_map_multiset_of[intro]:
  "a \<in># A \<Longrightarrow> f a \<in># image_mset f A"
  unfolding mem_set_of_iff[symmetric]
  unfolding set_of_image_mset by auto

definition Union_mset :: "'a multiset multiset \<Rightarrow> 'a multiset"
  where [code del]: "Union_mset \<equiv> Multiset.fold op + {#}"

lemma Union_mset_union[simp]: "Union_mset (M + N) = Union_mset M + Union_mset N"
  unfolding Union_mset_def by simp 

lemma Union_mset_empty[simp]: "Union_mset {#} = {#}" 
  unfolding Union_mset_def by simp

lemma Union_mset_singleton[simp]: "Union_mset {#M#} = M" 
  unfolding Union_mset_def by simp

lemma image_mset_Union_mset: "image_mset f (Union_mset Ms) = Union_mset (image_mset (image_mset f) Ms)"
  by (induct Ms, auto)

lemma Union_mset_mem_subset: assumes "M \<in># ms" shows "M \<le> Union_mset ms"
  using assms
proof (induct ms)
  case empty thus ?case by simp
next
  case (add ms N)
  show ?case
  proof (cases "M = N")
    case True
    thus ?thesis by auto
  next
    case False
    with add(2) have "M \<in># ms" by auto
    from add(1)[OF this] have "M \<le> Union_mset ms" .
    also have "... \<le> Union_mset ms + Union_mset {#N#}"
      by auto
    finally show ?thesis by simp
  qed
qed

lemma multiset_of_replicate_mono: assumes "m1 \<le> m2"
  shows "Union_mset (multiset_of (replicate n m1))
    \<le> Union_mset (multiset_of (replicate n m2))"
proof (induct n)
  case (Suc n)
  have "Union_mset (multiset_of (replicate (Suc n) m1)) = 
    Union_mset (multiset_of (replicate n m1)) + m1" by simp
  also have "\<dots> \<le> Union_mset (multiset_of (replicate n m1)) + m2" using `m1 \<le> m2` by auto
  also have "\<dots> \<le> Union_mset (multiset_of (replicate n m2)) + m2" using Suc by auto
  finally show ?case by auto
qed simp

lemma multised_add_singleton[simp]: "{# a #} + M = N + {# a #} \<longleftrightarrow> M = N"
  "M + {# a #} = {# a #} + N \<longleftrightarrow> M = N"
  by (auto simp: ac_simps)

lemma multiset_image_set_mono: assumes AB: "A \<le> B" shows "image_mset f A \<le> image_mset f B"
proof -
  from AB[unfolded mset_le_exists_conv] obtain C where B: "B = A + C" ..
  thus ?thesis by auto
qed

lemma mem_Union_mset[simp]: "x \<in># Union_mset ms \<longleftrightarrow> x \<in> \<Union> (set_of ` (set_of ms))"
  by (induct ms, auto)
  
lemma diff_union_swap_multiset :
  "a \<in># A \<Longrightarrow> A - {#a#} + B = A + B - {#a#}"
  by (intro multiset_eqI) auto

text{* In a multiset, there exists a local maxima. *}

lemma local_maxima_multiset:
  assumes "locally_trans s A A A" (is "?lt A")
    and "locally_irrefl s A" (is "?li A")
    and A_ne: "A \<noteq> {#}"
  shows "\<exists>a. a \<in># A \<and> (\<forall>b. b \<in># A \<longrightarrow> (b,a) \<notin> s)"
using assms
proof (induct A)
  case empty thus ?case by auto
next
  case (add A x)
  from lt_trans_l[OF `?lt (A + {# x #})`] have lt_A: "?lt A" .
  from li_trans_l[OF `?li (A + {# x #})`] have li_A: "?li A" .
  note IH = add(1)[OF lt_A li_A]
  show ?case
  proof (cases "A = {#}")
    case True
    with `?li (A + {# x #})`[unfolded locally_irrefl_def, rule_format] show ?thesis
    by simp
  next
    case False
    from IH[OF this]
    obtain a1 where
      a1_in_A: "a1 \<in># A"
      and a1_is_lm: "\<forall>b. b \<in># A \<longrightarrow> (b,a1) \<notin> s" by force
    show ?thesis
    proof (cases "(x,a1) \<in> s")
      case True note sxa1 = this
      have GL: "x \<in># A + {# x #}" by auto
      from add(3)[unfolded locally_irrefl_def]
      have GR1: "(x,x) \<notin> s" by simp
      {
        fix b assume b_in_A: "b \<in># A"
        have "(b,x) \<notin> s"
        proof (cases "(b,x) \<in> s")
          case True
          from b_in_A have b_in: "b \<in># A + {# x #}" by simp
          from a1_in_A have a1_in: "a1 \<in># A + {# x #}" by simp
          from b_in and a1_in and GL and True and sxa1 and
            add(3)[unfolded locally_irrefl_def] and
            add(2)[unfolded locally_trans_def]
          have "(b,a1) \<in> s"  by metis
          with a1_is_lm and b_in_A show ?thesis by auto
        qed 
      }
      hence GR2: "\<forall>b. b \<in># A \<longrightarrow> (b,x) \<notin> s" by simp
      from this and GL and GR1 show ?thesis by auto
    next
      case False
      from this and a1_is_lm
      have GR: "\<forall>b. b \<in># A + {# x #} \<longrightarrow> (b,a1) \<notin> s" by simp
      from a1_in_A have GL: "a1 \<in># A + {# x #}" by simp
      from GL and GR show ?thesis by metis
    qed
  qed
qed

lemma mult1_ab_a: "({#a#}, {#a,b#}) \<in> mult1 r"
  unfolding mult1_def
proof(rule, unfold split, intro exI conjI)
  show "{#a, b#} = {#a#} + {#b #}"
    by (simp add: ac_simps)
next
  show "{#a#} = {# a #} + {#}" by simp
qed simp

lemma mult1_ab_b: "({#b#}, {#a,b#}) \<in> mult1 r"
proof -
  have id: "{#a,b#} = {#b,a#}" by (simp add: ac_simps)
  show ?thesis unfolding id by (rule mult1_ab_a)
qed

declare mult1_union[intro!]

lemma mult1_mono_right[intro!]: assumes "(M,N) \<in> mult1 r"
  shows "((M + T, N + T) \<in> mult1 r)"
  using mult1_union[OF assms, of T] by (simp add: ac_simps)

lemma mult1_singleton[simp]: 
  shows "((M, {# t #}) \<in> mult1 r) = (\<forall> a. a \<in># M \<longrightarrow> (a,t) \<in> r)" (is "?l = ?r")
proof
  assume r: ?r
  show ?l
    unfolding mult1_def
    by (rule, unfold split, rule exI[of _ t], rule exI[of _ "{#}"], 
      rule exI[of _ M], insert r, auto)
next
  assume l: ?l
  from this[unfolded mult1_def]
  obtain a M0 K where id: "{#t#} = M0 + {#a#}"
    "M = M0 + K" and rel: "\<forall> b. b \<in># K \<longrightarrow> (b,a) \<in> r" by blast
  from id(1)[unfolded single_is_union] have "M0 = {#}" by simp
  with id have id: "t = a" "M = K" by auto
  show ?r unfolding id by (rule rel)
qed
end
