(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Order_Extension
imports
  Term_Order
  Signature_Extension
  Argument_Filter
  RPO
  Reduction_Pair
begin

context redtriple_order
begin

lemma sig_mono_imp_SN_rel:
  fixes R Rw :: "('f, 'v) trs"
  assumes sig_mono:
    "\<And> s t f bef aft. \<lbrakk>(s, t) \<in> S; (f, Suc (length bef + length aft)) \<in> F\<rbrakk> \<Longrightarrow>
    (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> S"
    and orient: "R \<subseteq> S" "Rw \<subseteq> NS"
    and var_cond: "\<And> l r. (l, r) \<in> NS \<inter> Rw \<Longrightarrow> vars_term r \<subseteq> vars_term l" 
    and F: "funas_trs R \<subseteq> F" "funas_trs Rw \<subseteq> F"
  shows "SN_rel (rstep R) (rstep Rw)"
proof -
  let ?R = "rstep R"
  let ?FR = "sig_step F ?R"
  let ?S = "rstep Rw"
  let ?FS = "sig_step F ?S"
  have S: "?FR \<subseteq> S"
  proof
    fix s t
    assume st: "(s, t) \<in> ?FR"
    hence st: "(s, t) \<in> ?R" and sF: "funas_term s \<subseteq> F" and tF: "funas_term t \<subseteq> F" using sig_stepE[OF st] by auto
    from st obtain C l r \<sigma> where lr: "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
    from orient lr have lr: "(l, r) \<in> S" by auto
    from subst_closedE[OF subst_S lr] have gt: "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S" .
    from sF[unfolded s] have "funas_ctxt C \<subseteq> F" by auto
    hence "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> S"
    proof (induction C)
      case Hole
      show ?case using gt by simp
    next
      case (More f bef C aft)
      from More.prems have C: "funas_ctxt C \<subseteq> F" and f: "(f, Suc (length bef + length aft)) \<in> F" by auto
      from sig_mono[OF More.IH[OF C] f] show ?case by simp
    qed
    thus st: "(s, t) \<in> S" unfolding s t .
  qed
  from trs_inclusion[OF wm_NS subst_NS orient(2)]
    have NS: "?S \<subseteq> NS" .
  hence NS_sig: "?FS \<subseteq> NS" by auto
  from relto_mono[OF S NS_sig]
    have subset: "relto (?FR) (?FS) \<subseteq> S" unfolding relto_S .
  have "SN (relto (?FR) (?FS))"
    by (rule SN_subset[OF SN subset])
  hence relSN: "SN_rel (?FR) (?FS)" unfolding SN_rel_defs .
  {
    fix l r
    assume lr: "(l, r) \<in> Rw"
    with orient have "(l, r) \<in> NS \<inter> Rw" by auto
    from var_cond[OF this] 
    have "vars_term r \<subseteq> vars_term l" .
  } note var_cond = this
  show ?thesis        
    by (rule sig_ext_relative_rewriting_var_cond[OF var_cond F relSN])
qed

lemma manna_ness_relative:
  assumes orient: "R \<subseteq> S" "Rw \<subseteq> NS"
    and mono: "ctxt_closed S"
  shows "SN_rel (rstep R) (rstep Rw)"
proof -
  from trs_inclusion[OF wm_NS subst_NS orient(2)] have NS: "rstep Rw \<subseteq> NS" .
  from trs_inclusion[OF mono subst_S orient(1)] have S: "rstep R \<subseteq> S" .
  from SN_subset[OF SN relto_mono[OF S NS, unfolded relto_S]]
  show ?thesis unfolding SN_rel_defs .
qed

lemma sig_mono_imp_SN_rel_ce: 
  assumes sig_mono:
    "\<And> s t f bef aft. \<lbrakk>(s, t) \<in> S; (f, Suc (length bef + length aft)) \<in> F\<rbrakk> \<Longrightarrow>
    (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> S"
    and F_unary: "\<And> fn. fn \<in> F \<Longrightarrow> snd fn \<le> Suc 0"
    and orient: "R \<subseteq> S" "Rw \<subseteq> NS"
    and var_cond: "\<And> l r. (l, r) \<in> NS \<inter> Rw \<Longrightarrow> vars_term r \<subseteq> vars_term l" 
    and F: "funas_trs R \<subseteq> F" "funas_trs Rw \<subseteq> F"
  shows "SN_rel (rstep (R \<union> \<Union>(ce_trs ` UNIV))) (rstep (Rw \<union> \<Union>(ce_trs ` UNIV)))"
proof -
  let ?c = "\<Union>(ce_trs ` UNIV) :: ('f, 'v) trs"
  from sig_mono_imp_SN_rel[OF sig_mono orient var_cond F]
    have SN_rel_R_Rw: "SN_rel (rstep R) (rstep Rw)" .
  hence SN: "SN_qrel (False, {}, R, Rw)" unfolding SN_qrel_def by simp
  have "SN_qrel (False, {}, R \<union> ?c, Rw \<union> ?c)"
  proof (rule SN_qrel_split[where D = ?c])
    show "SN_qrel (False, {}, R \<union> ?c - ?c, Rw \<union> ?c - ?c)"
      by (rule SN_qrel_mono[OF _ _ _ SN], auto)
  next
    show "SN_qrel (False, {}, ?c \<inter> (R \<union> ?c \<union> (Rw \<union> ?c)), R \<union> ?c \<union> (Rw \<union> ?c) - ?c)"
    proof (rule SN_qrel_mono[where Q = "{}" and Q' = "{}"], unfold SN_qrel_def split qrstep_rstep_conv)
      obtain pi :: "'f afs" where pi: "pi = (\<lambda> (f, n). if n = 0 then AFList [] else if n = Suc 0 then Collapse 0 else AFList [0 ..< n])" by auto
      obtain prc :: "'f filtered \<Rightarrow> nat \<Rightarrow> nat" where prc: "prc = (\<lambda> _ _. 0)"  by auto
      obtain c :: "'f filtered \<Rightarrow> nat \<Rightarrow> order_tag" where c: "c = (\<lambda> _ _ . Lex)" by auto
      obtain n :: nat where True by auto
      interpret rpo_pr "prc_nat prc" "prl_nat prc" c ..            
      let ?S = "RPO_S prc c n"
      let ?NS = "RPO_NS prc c n"
      interpret mono_ce_af_redtriple ?S ?NS ?NS full_af ..
      interpret afs_redtriple pi ?S ?NS ?NS
      proof
        show "wf_afs pi" unfolding wf_afs_def
        proof(intro allI)
          fix f n
          show "wf_af_entry n (pi (f, n))"
            unfolding pi
            by (cases n, simp, cases "n - 1", auto)
        qed
      qed
      let ?pi = "afs.afs_rel pi"
      show "SN_rel (rstep ?c) (rstep (R \<union> Rw))" 
      proof (rule manna_ness_relative[OF _ _ ctxt_closed[OF _ rpo_pr_prc.RPO_S_pr_ctxt_closed]]) 
        show "mono_afs pi" unfolding mono_afs_def
        proof (intro allI)
          fix f n
          show "mono_af_entry n (pi (f, n))" unfolding pi
            by (cases n, simp, cases "n - Suc 0", auto)
        qed
      next
        show "?c \<subseteq> ?pi ?S"
        proof
          fix s t
          assume st: "(s, t) \<in> ?c"
          then obtain c' d m where st: "(s, t) \<in> ce_trs (c', d, m)" by force
          show "(s, t) \<in> ?pi ?S" unfolding afs_rel_def
          proof(rule, rule exI, rule conjI[OF refl])            
            let ?s = "afs_term pi s"
            let ?t = "afs_term pi t"
            from st[unfolded ce_trs.simps]
            obtain ss where s: "s = Fun c' ss" and len: "length ss = Suc (Suc m)"
              and t: "t \<in> set ss" by auto
            have pi: "pi (c', length ss) = AFList [0 ..< length ss]"
              using len unfolding pi by simp
            have "?t \<in> set (args ?s)" unfolding s using t[unfolded set_conv_nth]
              by (force simp: Let_def pi) 
            hence "?s \<rhd> ?t" by (cases ?s, auto)
            from rpo_pr_prc.supt_imp_rpo_stri[OF this, of prc c n]
            have "(?s, ?t) \<in> ?S" unfolding rpo_pr_prc.RPO_S_pr_def by auto
            thus "afs_rule pi (s, t) \<in> ?S"
              unfolding afs_rule_def fst_conv snd_conv .
          qed
        qed
      next
        show "R \<union> Rw \<subseteq> afs.afs_rel pi ?NS"
        proof
          fix l r
          assume lr: "(l, r) \<in> R \<union> Rw"
          with F have F: "funas_term l \<subseteq> F" "funas_term r \<subseteq> F"
            unfolding funas_trs_def funas_rule_def [abs_def] by force+
          from SN_rel_imp_wf_reltrs[OF SN_rel_R_Rw]
          have "wf_reltrs (R) (Rw)" .
          with var_cond[of l r] lr orient(2) have vars: "vars_term r \<subseteq> vars_term l"
            unfolding wf_reltrs.simps wf_trs_def by auto
          let ?pi = "afs_term pi"
          { fix t :: "('f, 'v) term" and x
            assume "funas_term t \<subseteq> F" and "x \<in> vars_term t"
            hence "?pi t = Var x"
            proof (induction t)
              case (Var y)
              thus ?case by simp
            next
              case (Fun f ts)
              from Fun.prems(1) have "(f, length ts) \<in> F" by auto
              from F_unary[OF this] obtain t where ts: "ts = [t]"
                using Fun.prems(2) by (cases ts, auto)
              from Fun.prems[unfolded ts] have F: "funas_term t \<subseteq> F"
                and x: "x \<in> vars_term t" by auto
              from Fun.IH[OF _ F x] show ?case unfolding ts 
                unfolding pi by simp
            qed }
          note var = this
          have "(?pi l, ?pi r) \<in> ?NS"
          proof (cases "vars_term r = {}")
            case False
            then obtain x where xr: "x \<in> vars_term r" by auto
            with vars have xl: "x \<in> vars_term l" by auto
            show ?thesis
              unfolding var[OF F(1) xl] var[OF F(2) xr]
              by (rule rpo_pr_prc.RPO_refl)
          next
            case True
            with F(2) have "\<exists> c. ?pi r = Fun c []"
            proof (induction r)
              case (Var x) thus ?case by auto
            next
              case (Fun f ts)
              from Fun.prems(1) have "(f, length ts) \<in> F" by auto
              from F_unary[OF this] obtain t where ts: "ts = [t] \<or> ts = []"
                using Fun.prems(2) by (cases ts, auto)
              thus ?case
              proof
                assume ts: "ts = [t]"
                from Fun.prems[unfolded ts] have F: "funas_term t \<subseteq> F"
                  and vars: "vars_term t = {}" by auto
                from Fun.IH[OF _ F vars] show ?case unfolding ts 
                  unfolding pi by simp
              next
                assume ts: "ts = []"
                show ?case unfolding ts unfolding pi by simp
              qed
            qed
            then obtain f where pir: "?pi r = Fun f []" by auto
            have f: "prl_nat prc f" unfolding prc prl_nat_def by simp
            show ?thesis unfolding pir 
              using rpo_pr_prc.RPO_least_Var[OF f]
              using rpo_pr_prc.RPO_least_Fun[OF f]
              by (cases "?pi l", auto)
          qed
          thus "(l, r) \<in> afs.afs_rel pi ?NS" 
            unfolding afs_rel_def afs_rule_def by auto
        qed
      qed
    qed auto
  qed
  thus ?thesis unfolding SN_qrel_def by simp
qed

end

lemma ce_SN_rel_imp_redtriple:
  fixes R Rw :: "('f, 'v) trs"
  assumes SN_rel: "SN_rel (rstep R) (rstep Rw)"
    and ce: "\<Union>(ce_trs ` UNIV) \<subseteq> R \<inter> Rw"
    and pi: "\<And> f n i. i < n \<Longrightarrow> pi (f, n) i"
  shows "mono_ce_af_redtriple_order ((relto (rstep R) (rstep Rw))\<^sup>+) ((rstep Rw)\<^sup>*) ((rstep Rw)\<^sup>*) pi"
proof -
  let ?R = "relto (rstep R) (rstep Rw)"
  let ?Rw = "rstep Rw"
  from SN_rel have SN: "SN (?R^+)" unfolding SN_rel_defs
    by (rule SN_imp_SN_trancl)
  note conv = relto_trancl_conv
  note subst = subst_closed_rtrancl subst_closed_union subst_closed_comp subst_closed_rstep
  let ?ce = "\<Union>(range ce_trs) :: ('f, 'v)trs"
  from ce have "?ce \<subseteq> rstep R \<inter> rstep Rw" by auto
  also have "... \<subseteq> ?R^+ \<inter> ?Rw^*" unfolding conv by auto
  finally have ce: "?ce \<subseteq> ?R^+ \<inter> ?Rw^*" .
  {
    fix c d m
    have "ce_trs (c, d, m) \<subseteq> ?ce" by auto
    also have "... \<subseteq> ?R^+ \<inter> ?Rw^*" by (rule ce)
    finally have "ce_trs (c, d, m) \<subseteq> ?R^+ \<inter> ?Rw^*" .
  } note ce = this
  show ?thesis
  proof(unfold_locales, rule SN)
    show "ctxt_closed (?Rw^*)" by blast
    show "subst_closed (?R^+)" unfolding conv
      by (blast intro: subst)
    show "subst_closed (?Rw^*)" by (blast intro: subst)
    show "refl (?Rw^*)" by auto
    show "trans (?Rw^*)" by auto
    show "?Rw^* O ?R^+ \<subseteq> ?R^+" unfolding conv by regexp
    show "?R^+ O ?Rw^* \<subseteq> ?R^+" unfolding conv by regexp
    show "af_compatible pi (?Rw^*)" unfolding af_compatible_def  using pi by auto
    have "?R^+ O ?R^+ \<subseteq> ?R^+" unfolding conv by regexp
    thus "trans (?R^+)" by auto    
  qed (insert ce, blast+)
qed

end

